from datetime import datetime
from time import sleep


def retry(func=None, *, timeout=None, delay=None):
    def _decorate(function):
        def wrapped_function(*args, **kwargs):
            start_time = datetime.now()
            if not isinstance(timeout, (int, float)):
                raise ValueError(
                    "timeout: expecting int/float but was {}".format(
                        type(timeout).__name__
                    )
                )
            if not isinstance(delay, (int, float)):
                raise ValueError(
                    "delay: expecting int/float but was {}".format(type(delay).__name__)
                )
            while function(*args, **kwargs) is False:
                if (datetime.now() - start_time).total_seconds() <= timeout:
                    sleep(delay)
                else:
                    return False
            return True

        return wrapped_function

    if func:
        return _decorate(func)
    return _decorate


class TmuxOutput(object):
    def __init__(self, func, timeout=None, delay=None):
        self.func = func
        self.value = self.func()
        self.__timeout = timeout
        self.__delay = delay

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)

    def __eq__(self, other):
        @retry(timeout=self.__timeout, delay=self.__delay)
        def _test():
            self.value = self.func()
            return self.value == other

        return _test()

    def __ne__(self, other):
        @retry(timeout=self.__timeout, delay=self.__delay)
        def _test():
            self.value = self.func()
            return not self.value == other

        return _test()

    def __contains__(self, other):
        @retry(timeout=self.__timeout, delay=self.__delay)
        def _test():
            self.value = self.func()
            return other in self.value

        return _test()



from inspect import cleandoc

from .output import TmuxOutput


class Client:
    __allowed_params = [
        "session_name",
        "start_directory",
        "window_name",
        "window_command",
        "window_height",
        "window_width",
        "assert_timeout",
        "assert_delay",
        "debug",
    ]

    @classmethod
    def list_params(cls):
        return cls.__allowed_params

    def __init__(self, tmux_server=None, pytest_request=None, **kwargs):
        self.pytest_request = pytest_request
        self._config = {}
        for x in self.__class__.__allowed_params:
            self._config[x] = kwargs.get(x, None)

        self._server = tmux_server
        self._session = None
        self._window = None
        self._pane = None
        self._config["session_name"] = pytest_request.module.__name__
        self._config["session_name"] = self._config["session_name"].replace(".", " ")
        self._config["session_name"] = self._config["session_name"].replace(":", " ")
        self._config["window_name"] = self.pytest_request.node.name
        self._previous_config = dict(self._config)

    class suspend_capture:
        def __init__(self, pytestconfig=None):
            self.capmanager = pytestconfig.pluginmanager.getplugin("capturemanager")

        def __enter__(self):
            self.capmanager.suspend_global_capture(in_=True)

        def __exit__(self, _1, _2, _3):
            try:
                input("Press enter to continue....")
            except KeyboardInterrupt:
                self.capmanager.resume_global_capture()

    def __debug(self, msg=None):

        if self.config["debug"]:
            with self.suspend_capture(self.pytest_request.config):
                print("")
                print(cleandoc(msg))

    @property
    def config(self):
        return self._config

    @property
    def previous_config(self):
        return self._previous_config

    def set(self, **kwargs):
        self._previous_config = dict(self._config)
        self._config.update(kwargs)

    def restore_previous_config(self):
        self._config = dict(self._previous_config)

    @property
    def session(self):
        if self._session is None:
            self.__debug("PyTestTmuxPlugin started with DEBUG")

            self._session = self.server.new_session(
                session_name=self.config["session_name"],
                window_name=self.config["window_name"],
                window_command=self.config["window_command"],
            )

            if self.config["window_width"]:
                self.window.set_window_option(
                    "force-width", self.config["window_width"]
                )

            if self.config["window_height"]:
                self.window.set_window_option(
                    "force-height", self.config["window_height"]
                )

            self.__debug(
                """
                        Open a new window terminal and use the bellow command to connect to the tmux session
                        tmux -S "{}" attach -t "{}"
                        """.format(
                    self.server.socket_path, self.config["session_name"]
                )
            )

        return self._session

    @property
    def window(self):
        if self._window is None:
            self._window = self.session.attached_window
        return self._window

    @property
    def pane(self):
        if self._pane is None:
            self._pane = self.window.attached_pane
        return self._pane

    @property
    def server(self):
        return self._server

    def kill_session(self):
        self.__debug(
            """
                    killing the session "{}"
                    """.format(
                self.session._info["session_name"]
            )
        )
        self.session.kill_session()

    def clear(self):
        self.pane.clear()

    def send_keys(self, cmd=None, **kwargs):
        if "supress_history" not in kwargs:
            kwargs["suppress_history"] = False
        self.__debug(
            """
                    Send "{}" to tmux session
                    """.format(
                cmd
            )
        )
        self.pane.send_keys(cmd, **kwargs)

    def screen(self, timeout=None, delay=None):
        if timeout is None:
            timeout = self.config["assert_timeout"]
        if delay is None:
            delay = self.config["assert_delay"]

        def _capture():
            return "\n".join(self.pane.capture_pane())

        return TmuxOutput(_capture, timeout=timeout, delay=delay)

    def row(self, row, timeout=None, delay=None):
        if not isinstance(row, int):
            raise TypeError("row should be an integer")

        if timeout is None:
            timeout = self.config["assert_timeout"]
        if delay is None:
            delay = self.config["assert_delay"]

        def _capture():
            try:
                output = self.pane.capture_pane()[row]
            except IndexError:
                output = ""
            return output

        return TmuxOutput(_capture, timeout=timeout, delay=delay)



#!/usr/bin/env python#
import os

from libtmux import Server

from pytest_tmux.client import Client
from pytest_tmux.fixtures import tmux, tmux_server
from pytest_tmux.rewrite import tmux_rewrite

(tmux_server, tmux)


def pytest_addoption(parser):
    """Add options to control tmux tests"""
    group = parser.getgroup("tmux")
    group.addoption(
        "--tmux-debug",
        dest="tmux_debug",
        action="store_true",
        default=os.getenv("PYTEST_TMUX_DEBUG", False) in ("True", "1"),
        help="""
            Prompt the user to press enter during tests who use tmux ficture.
            Env: PYTEST_TMUX_DEBUG
        """,
    )
    group.addoption(
        "--tmux-socket-path",
        dest="tmux_socket_path",
        action="store",
        default=os.getenv("PYTEST_TMUX_SOCKET_PATH", None),
        help="""
            Socket path to use with libtmux.Server
            Default: [pytest_temp_dir]/tmux.socket
            Env: PYTEST_TMUX_SOCKET_PATH
        """,
    )
    group.addoption(
        "--tmux-config-file",
        dest="tmux_config_file",
        action="store",
        default=os.getenv("PYTEST_TMUX_CONFIG_FILE", None),
        help="""
            Path to a tmux configuration file to use with libtmux.Server
            Default: None
            Env: PYTEST_TMUX_CONFIG_FILE
        """,
    )
    group.addoption(
        "--tmux-colors",
        dest="tmux_colors",
        type=int,
        action="store",
        default=os.getenv("PYTEST_TMUX_COLORS", None),
        help="""
            Number of colors to use with libtmux.Server
            Default: None
            Env: PYTEST_TMUX_COLORS
        """,
    )
    group.addoption(
        "--tmux-start-directory",
        dest="tmux_start_directory",
        action="store",
        default=os.getenv("PYTEST_TMUX_START_DIRECTORY", None),
        help="""
            Working directory in which the new window is created
            Default: None
            Env: PYTEST_TMUX_START_DIRECTORY
        """,
    )
    group.addoption(
        "--tmux-window-command",
        dest="tmux_window_command",
        action="store",
        default=os.getenv("PYTEST_TMUX_WINDOW_COMMAND", None),
        help="""
            Command executed when starting the session.
            Default: None
            Env: PYTEST_TMUX_WINDOW_COMMAND
        """,
    )
    group.addoption(
        "--tmux-window-width",
        dest="tmux_window_width",
        type=int,
        action="store",
        default=os.getenv("PYTEST_TMUX_WINDOW_WIDTH", None),
        help="""
            Force tmux window width
            Default: None
            Env: PYTEST_TMUX_WINDOW_WIDTH
        """,
    )
    group.addoption(
        "--tmux-window-height",
        dest="tmux_window_height",
        type=int,
        action="store",
        default=os.getenv("PYTEST_TMUX_WINDOW_HEIGHT", None),
        help="""
            Force tmux window height
            Default: None
            Env: PYTEST_TMUX_WINDOW_HEIGHT
        """,
    )
    group.addoption(
        "--tmux-assert-timeout",
        dest="tmux_assert_timeout",
        type=int,
        action="store",
        default=os.getenv("PYTEST_TMUX_ASSERT_TIMEOUT", 2),
        help="""
            Seconds before tmux assertion should fail
            Default: 2
            Env: PYTEST_TMUX_ASSERT_TIMEOUT
        """,
    )
    group.addoption(
        "--tmux-assert-delay",
        dest="tmux_assert_delay",
        type=int,
        action="store",
        default=os.getenv("PYTEST_TMUX_ASSERT_DELAY", 0.5),
        help="""
            Seconds to wait before tmux assertion should retry
            Default: 0.5
            Env: PYTEST_TMUX_ASSERT_DELAY
        """,
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "tmux_cfg(**cfg_opts): Allow change tmux fixtures options for the current test",
    )
    assert config.pluginmanager.register(PyTestTmuxPlugin(config), "tmux")


class PyTestTmuxPlugin:
    def __init__(self, config):
        self.config = config

    def getconfig(self, config=None, request=None, **kwargs):
        tmux_cfg = {}
        """ Global/Session config """
        if config is not None:
            tmux_cfg.update(self._load_tmux_config(config))
        else:
            tmux_cfg.update(self._load_tmux_config(self.config))
        """ Test / Marker config """
        if request is not None:
            tmux_cfg.update(self._load_markers_config(request))
        tmux_cfg.update(kwargs)

        """
        Set defaults who can't be set by pytest_addoption
        """
        if tmux_cfg["socket_path"] is None:
            tmux_cfg["socket_path"] = (
                self.config._tmpdirhandler.getbasetemp() + "/tmux.socket"
            )

        return tmux_cfg

    def _load_tmux_config(self, config):
        """Load configuration from '--tmux-*' command-line."""
        cfg = {}
        for k, v in vars(config.option).items():
            if k.startswith("tmux_"):
                short_key = k[5:]
                cfg[short_key] = v

        return cfg

    def _load_markers_config(self, request):
        """Load configuration from 'tmux_cfg' markers."""
        cfg = {}

        marker = request.node.get_closest_marker("tmux_cfg")
        if marker:
            cfg = marker.kwargs

        return cfg

    def server(self):
        """
        Return a libtmux.Server object (use by tmux_server session fixture)
        """
        tmux_server_cfg = {}
        tmux_cfg = self.getconfig()
        tmux_server_keys = ["socket_path", "config_file", "colors"]
        for k, v in tmux_cfg.items():
            if k in tmux_server_keys:
                if v is not None:
                    tmux_server_cfg[k] = v

        return Server(**tmux_server_cfg)

    def client(self, tmux_server, pytest_request, **kwargs):
        """
        Return a pytest_tmux.Client object (used by tmux function fixture)
        """

        if not isinstance(tmux_server, Server):
            raise ValueError("First arg need to be a Server instance")

        if type(pytest_request).__name__ != "SubRequest":
            raise ValueError("Second argument need to be a SubRequest instance")

        tmux_client_cfg = {}
        tmux_cfg = kwargs
        """
        Retrieve allowed parameters for Client class
        Then filtering config parameters to use
        """
        tmux_client_keys = Client.list_params()

        for k, v in tmux_cfg.items():
            if k in tmux_client_keys:
                if v is not None:
                    tmux_client_cfg[k] = v

        return Client(
            tmux_server=tmux_server, pytest_request=pytest_request, **tmux_client_cfg
        )


def pytest_assertrepr_compare(op, left, right):
    if type(left).__name__ == "TmuxOutput" or type(right).__name__ == "TmuxOutput":
        return tmux_rewrite(op, left, right)



import pytest


@pytest.fixture(scope="session")
def tmux_server(request):
    plugin = request.config.pluginmanager.getplugin("tmux")
    tmux_server = plugin.server()

    yield tmux_server

    if tmux_server:
        tmux_server.kill_server()


@pytest.fixture()
def tmux(request, tmux_server):
    plugin = request.config.pluginmanager.getplugin("tmux")
    tmux_cfg = plugin.getconfig(request=request)
    tmux_client = plugin.client(tmux_server, request, **tmux_cfg)

    yield tmux_client

    if tmux_client:
        if tmux_client.session:
            tmux_client.kill_session()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_tmux']

package_data = \
{'': ['*']}

install_requires = \
['libtmux==0.16']

entry_points = \
{'pytest11': ['pytest-tmux = pytest_tmux.plugin']}

setup_kwargs = {
    'name': 'pytest-tmux',
    'version': '0.0.1',
    'description': 'A pytest plugin that enables tmux driven tests',
    'long_description': '# pytest-tmux\n\n[![PyPI version](https://img.shields.io/pypi/v/pytest-tmux.svg)](https://pypi.org/project/pytest-tmux)\n\n[![Python versions](https://img.shields.io/pypi/pyversions/pytest-tmux.svg)](https://pypi.org/project/pytest-tmux)\n\n[![See Build Status on AppVeyor](https://ci.appveyor.com/api/projects/status/github/rockandska/pytest-tmux?branch=master)](https://ci.appveyor.com/project/rockandska/pytest-tmux/branch/master)\n\nA pytest plugin that enables tmux driven tests\n\n------------------------------------------------------------------------\n\n## Features\n\n- Enable tmux driven tests\n- Enable screen assertion\n- Enable row assertion\n\n## Requirements\n\n- python >= 3.7\n- python libtmux==0.16\n- pytest\n- tmux\n\n## Installation\n\nYou can install "pytest-tmux" via [pip](https://pypi.org/project/pip/)\nfrom [PyPI](https://pypi.org/project):\n\n    $ pip install pytest-tmux\n\n## Purpose and design\n\nThis plugin is intend to help users who whant to test interrative cli.\nWhen using `tmux` fixture it basically :\n- create a tmux server (socket create in tmux tmpdir)\n- create a session automatically when requested with name based on the name of test file\n- create a window automatically when requested with name based on the name of the test\n\nConfiguration could be set on different level :\n- cli args (see --tmux-* switch with pytest --help)\n- at the test level with tmux_cfg marker\n- dynamically inside test with `tmux.set()`\n- using `tmux.screen() / tmux.row()` (timeout / delay)\n\n\n## Usage\n\n### Basic example\n\n#### Success\n\n#### Code\n```python\nimport pytest\nfrom inspect import cleandoc\n\ndef test_assert(tmux):\n    # Set some options before session / windows is started\n    tmux.set(window_command=\'env -i PS1="$ " TERM="xterm-256color" /usr/bin/env bash --norc --noprofile\')\n    assert tmux.screen() == \'$\'\n    tmux.send_keys(r\'printf "  Hello World  .\\n\\n"\')\n    expected=r"""\n    $ printf "  Hello World  .\\n\\n"\n      Hello World  .\n\n    $\n    """\n    assert tmux.screen() == cleandoc(expected)\n```\n\n### Failure\n\n#### Code\n```python\nimport pytest\nfrom inspect import cleandoc\n\ndef test_assert(tmux):\n    # Set some options before session / windows is started\n    tmux.set(window_command=\'env -i PS1="# " TERM="xterm-256color" /usr/bin/env bash --norc --noprofile\')\n    assert tmux.row(0) == \'$\'\n```\n\n#### Output\n\n```\n>       assert tmux.row(0) == \'$\'\nE       assert failed\nE         > Common line\nE         - Left\nE         + Right\nE         -------------\nE         - #\nE         + $\nE         -------------\n```\n\n### Waiting for a long process\n\n```python\nimport pytest\nfrom inspect import cleandoc\n\ndef test_assert(tmux):\n    # Set some options before session / windows is started\n    tmux.set(window_command=\'env -i PS1="$ " TERM="xterm-256color" /usr/bin/env bash --norc --noprofile\')\n    assert tmux.row(0) == \'$\'\n    tmux.send_keys(\'sleep 5\')\n    assert tmux.row(0) == \'$ sleep 5\'\n    expected = """\n    $ sleep 5\n    $\n    """\n    assert tmux.screen(timeout=6, delay=0.5) == cleandoc(expected)\n```\n\n### Debug\n\nIf needed, a debug mode is available with `--tmux-debug`.\nIt will prompt you to :\n- open the tmux session for the current test\n- press enter to continue on :\n    - send_keys\n    - kill_session\n\n## Contributing\n\nContributions are very welcome.\n\n### Dev requirements\n\n- minor python3 versions present in .python-version\n- pyenv > v2.3.9 ( optional but recommended )\n- virtualenv (will be installed by `make venv` if not available)\n\n### Start hacking\n\n#### Install python versions used for tests ( recommended )\n\n```shell\npyenv install\n```\n\n/!\\ Install multiple versions in a single command only available in pyenv > 2.3.9\n\n#### Create the project virtualenv\n\n```shell\n$ make venv\n```\n\nThis target does the following:\n\n- installs `virtualenv` if not found\n- creates a venv in `.venv` for `dev-requirements.txt`\n- installs `dev-requirements.txt`\n- creates a `.python-venv` symlink to `.venv/bin/activate`\n\n### Run tests\n\nTests are driven by tox / make\n\n#### Run with make\n\nExample:\n```shell\n$ make test\n```\n\n#### Run with tox or free commands\n\nExample:\n```shell\n$ source .python-venv\n$ tox\n$ poetry run pytest tests/\n```\n\n#### Formater\n\n`black` and `isort` run in check mode by default\n\ncheck mode could be removed by running the env as :\n```shell\n$ source .python-venv\n$ tox -e black --\n$ tox -e isort --\n```\n\n### Update tox.ini / pyproject.toml\n\n- Update `tox.ini.j2` to update `tox.ini`\n- Update `pyproject.ini.j2` to update `pyproject.ini`\n- Update `.python-version` to add a new python version (order is important)\n\n## License\n\nDistributed under the terms of the\n[MIT](http://opensource.org/licenses/MIT) license, "pytest-tmux" is free\nand open source software\n\n## Issues\n\nIf you encounter any problems, please [file an\nissue](https://github.com/rockandska/pytest-tmux/issues) along with a\ndetailed description.\n',
    'author': 'rockandska',
    'author_email': 'yoann_mac_donald@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<=3.11.9999',
}


setup(**setup_kwargs)



def tmux_rewrite(op, left, right):
    diff = []
    left_arr = str(left).split("\n")
    right_arr = str(right).split("\n")
    if op == "==":
        diff.append("failed")
        diff.append("> Common line")
        diff.append("- Left")
        diff.append("+ Right")
        diff.append("-------------")
        for i in range(len(max([left_arr, right_arr], key=len))):
            try:
                left = left_arr[i]
            except IndexError:
                left = None

            try:
                right = right_arr[i]
            except IndexError:
                right = None

            if op == "==":
                if left != right:
                    if left is not None:
                        diff.append(f"- {left}")
                    if right is not None:
                        diff.append(f"+ {right}")
                else:
                    diff.append(f"> {left}")
        diff.append("-------------")
    elif op == "!=":
        diff.append("failed")
        diff.append("left and right are equal")
        diff.append("-------------")
        diff.extend(left_arr)
        diff.append("-------------")
    elif op == "in":
        diff.append("failed")
        diff.append("-------------")
        diff.extend(left_arr)
        diff.append("-------------")
        diff.append("was not found in")
        diff.append("-------------")
        diff.extend(right_arr)
        diff.append("-------------")
    return diff
